/* Snark - Main snark program startup class.
   Copyright (C) 2003 Mark J. Wielaard

   This file is part of Snark.
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package org.klomp.snark;

import java.io.*;
import java.net.*;
import java.util.*;

import org.klomp.snark.bencode.*;

/**
 * Main Snark program startup class.
 *
 * @author Mark Wielaard (mark@klomp.org)
 */
public class Snark
{
  private final static int MIN_PORT = 6881;
  private final static int MAX_PORT = 6889;

  // Error messages (non-fatal)
  public final static int ERROR   = 1;

  // Warning messages
  public final static int WARNING = 2;

  // Notices (peer level)
  public final static int NOTICE  = 3;

  // Info messages (protocol policy level)
  public final static int INFO    = 4;

  // Debug info (protocol level)
  public final static int DEBUG   = 5;

  // Very low level stuff (network level)
  public final static int ALL     = 6;

  /**
   * What level of debug info to show.
   */
  public static int debug = NOTICE;

  private static final String newline = System.getProperty("line.separator");

  private static final String copyright =
  "The Hunting of the Snark Project - Copyright (C) 2003 Mark J. Wielaard"
  + newline + newline
  + "Snark comes with ABSOLUTELY NO WARRANTY.  This is free software, and"
  + newline
  + "you are welcome to redistribute it under certain conditions; read the"
  + newline
  + "COPYING file for details.";
  
  private static final String usage =
  "Press return for help. Type \"quit\" and return to stop";
  
  public static void main(String[] args)
  {
    System.out.println(copyright);
    System.out.println();

    // Parse debug, share/ip and torrent file options.
    Snark snark = parseArguments(args);

    SnarkShutdown snarkhook
      = new SnarkShutdown(snark.storage,
			  snark.coordinator,
			  snark.acceptor,
			  snark.trackerclient);
    Runtime.getRuntime().addShutdownHook(snarkhook);

    boolean quit = false;
    
    System.out.println();
    System.out.println(usage);
    System.out.println();

    try
      {
	BufferedReader br = new BufferedReader
	  (new InputStreamReader(System.in));
	String line = br.readLine();
	while(!quit && line != null)
	  {
	    line = line.toLowerCase();
	    if ("quit".equals(line))
	      quit = true;
	    else if ("".equals(line) || "help".equals(line))
		System.out.println(usage);
	    else
	      {
		System.out.println("Unknown command: " + line);
		System.out.println(usage);
	      }
	    System.out.println();
	    line = br.readLine();
	  }
      }
    catch(IOException ioe)
      {
	debug("ERROR while reading stdin: " + ioe, ERROR);
      }
    
    Runtime.getRuntime().removeShutdownHook(snarkhook);
    snarkhook.run();
    System.exit(0);
  }

  Storage storage;
  PeerCoordinator coordinator;
  ConnectionAcceptor acceptor;
  TrackerClient trackerclient;

  private Snark(String torrent, String ip, int user_port)
  {

    // "Taking Three as the subject to reason about--
    // A convenient number to state--
    // We add Seven, and Ten, and then multiply out
    // By One Thousand diminished by Eight.
    //
    // "The result we proceed to divide, as you see,
    // By Nine Hundred and Ninety Two:
    // Then subtract Seventeen, and the answer must be
    // Exactly and perfectly true.

    // Create a new ID and fill it with something random.  First nine
    // zeros bytes, then three bytes filled with three and then
    // sixteen random bytes.
    byte[] id = new byte[20];
    Random random = new Random();
    int i;
    for (i = 0; i < 9; i++)
      id[i] = 0;
    id[i++] = 3;
    id[i++] = 3;
    id[i++] = 3;
    while (i < 20)
      id[i++] = (byte)random.nextInt(256);

    int port;
    IOException lastException = null;
    ServerSocket serversocket = null;
    if (user_port != -1)
      {
	port = user_port;
	try
	  {
	    serversocket = new ServerSocket(port);
	  }
	catch(IOException ioe)
	  {
	    lastException = ioe;
	  }
      }
    else
      {
	for (port = MIN_PORT; serversocket == null && port <= MAX_PORT; port++)
	  {
	    try
	      {
		serversocket = new ServerSocket(port);
	      }
	    catch(IOException ioe)
	      {
		lastException = ioe;
	      }
	  }
      }
    if (serversocket == null)
      {
	String message = "Cannot accept incoming connections ";
	if (user_port == -1)
	  message = message + "tried ports " + MIN_PORT + " - " + MAX_PORT;
	else
	  message = message + "on port " + user_port ;
	
	if (ip != null || user_port != -1)
	  fatal(message, lastException);
	else
	  debug("WARNING: " + message, WARNING);
	port = -1;
      }
    else
      {
	port = serversocket.getLocalPort();
	debug("Listening on port: " + port, Snark.NOTICE);
      }

    // Figure out what the torrent argument represents.
    MetaInfo meta = null;
    File f = null;
    try
      {
	InputStream in;
	f = new File(torrent);
	if (f.exists())
	  in = new FileInputStream(f);
	else
	  {
	    URL u = new URL(torrent);
	    URLConnection c = u.openConnection();
	    c.connect();
	    in = c.getInputStream();

	    if (c instanceof HttpURLConnection)
	      {
		// Check whether the page exists
		int code = ((HttpURLConnection)c).getResponseCode();
		if (code/100 != 2) // We can only handle 200 OK responses
		  fatal("Loading page '" + torrent + "' gave error code "
			+ code + ", it probably doesn't exists");
	      }
	  }
	meta = new MetaInfo(new BDecoder(in));
      }
    catch(IOException ioe)
      {
	// OK, so it wasn't a torrent metainfo file.
	if (f != null && f.exists())
	  if (ip == null)
	    fatal("'" + torrent + "' exists,"
		  + " but is not a valid torrent metainfo file."
		  + System.getProperty("line.separator")
		  + "  (use --share to create a torrent from it"
		  + " and start sharing)", ioe);
		 else
	    {
	      // Try to create a new metainfo file
	     Snark.debug
	       ("Trying to create metainfo torrent for '" + torrent + "'",
		NOTICE);
	     try
	       {
		 storage = new Storage
		   (f, "http://" + ip + ":" + port + "/announce");
		 meta = storage.getMetaInfo();
	       }
	     catch (IOException ioe2)
	       {
		 fatal("Could not create torrent for '" + torrent + "'", ioe2);
	       }
	    }
	else
	  fatal("Cannot open '" + torrent + "'", ioe);
      }
    
    debug(meta.toString(), INFO);

    // When the metainfo torrent was created from an existing file/dir
    // it alreadt exists.
    if (storage == null)
      {
	try
	  {
	    storage = new Storage(meta);
	  }
	catch (IOException ioe)
	  {
	    fatal("Could not create storage", ioe);
	  }
      }

    coordinator = new PeerCoordinator(id, meta, storage);
    HttpAcceptor httpacceptor;
    if (ip != null)
      {
	MetaInfo m = meta.reannounce
	  ("http://" + ip + ":" + port + "/announce");
	Tracker tracker = new Tracker(m);
	try
	  {
	    tracker.addPeer(new PeerID(id, InetAddress.getByName(ip), port));
	  }
	catch (UnknownHostException oops)
	  {
	    fatal("Could not start tracker for " + ip, oops);
	  }
	httpacceptor = new HttpAcceptor(tracker);
      }
    else
      httpacceptor = null;

    PeerAcceptor peeracceptor = new PeerAcceptor(coordinator);
    ConnectionAcceptor acceptor = new ConnectionAcceptor(serversocket,
							 httpacceptor,
							 peeracceptor);

    if (ip != null)
      Snark.debug("Torrent available on "
		  + "http://" + ip + ":" + port
		  + "/metainfo.torrent",
		  NOTICE);

    trackerclient = new TrackerClient(meta, coordinator, port);
    trackerclient.start();

  }

  /**
   * Sets debug, ip and torrent variables then creates a Snark instance.
   * Calls usage(), which terminates the program, if non-valid argument list.
   */
  private static Snark parseArguments(String[] args)
  {
    int user_port = -1;
    String ip = null;
    String torrent = null;

    int i = 0;
    while (i < args.length)
      {
	if (args[i].equals("--debug"))
	  {
	    debug = INFO;
	    i++;

	    // Try if there is an level argument.
	    if (i < args.length)
	      {
		try
		  {
		    int level = Integer.parseInt(args[i]);
		    if (level >= 0)
		      {
			debug = level;
			i++;
		      }
		  }
		catch (NumberFormatException nfe) { }
	      }
	  }
	else if (args[i].equals("--port"))
	  {
	    if (args.length - 1 < i + 1)
	      usage("--port needs port number to listen on");
	    try
	      {
		user_port = Integer.parseInt(args[i + 1]);
	      }
	    catch (NumberFormatException nfe)
	      {
		usage("--port argument must be a number (" + nfe + ")");
	      }
	    i += 2;
	  }
	else if (args[i].equals("--share"))
	  {
	    if (args.length - 1 < i + 1)
	      usage("--share needs local ip-address or host-name");
	    ip = args[i + 1];
	    i += 2;
	  }
	else
	  {
	    torrent = args[i];
	    i++;
	    break;
	  }
      }

    if (torrent == null || i != args.length)
      if (torrent != null && torrent.startsWith("-"))
	usage("Unknow option '" + torrent + "'.");
      else
	usage("Need exactly one <url>, <file> or <dir>.");

    return new Snark(torrent, ip, user_port);
  }
  
  private static void usage(String s)
  {
    System.out.println("snark: " + s);
    usage();
  }

  private static void usage()
  {
    System.out.println
      ("Usage: snark [--debug [level]] [--port <port>] [--share (<ip>|<host>)]");
    System.out.println
      ("             (<url>|<file>|<dir>)");
    System.out.println
      ("  --debug\tShows some extra info and stacktraces");
    System.out.println
      ("    level\tHow much debug details to show");
    System.out.println
      ("         \t(defaults to "
       + NOTICE + ", with --debug to "
       + INFO + ", highest level is "
       + ALL + ").");
    System.out.println
      ("  --port\tThe port to listen on for incomming connections");
    System.out.println
      ("        \t(if not given defaults to first free port between "
       + MIN_PORT + "-" + MAX_PORT + ").");
    System.out.println
      ("  --share\tStart torrent tracker on <ip> address or <host> name.");
    System.out.println
      ("  <url>  \tURL pointing to .torrent metainfo file to download/share.");
    System.out.println
      ("  <file> \tEither a local .torrent metainfo file to download");
    System.out.println
      ("         \tor (with --share) a file to share.");
    System.out.println
      ("  <dir>  \tA directory with files to share (needs --share).");
    System.exit(-1);
  }

  /**
   * Aborts program abnormally.
   */
  public static void fatal(String s)
  {
    fatal(s, null);
  }

  /**
   * Aborts program abnormally.
   */
  public static void fatal(String s, Throwable t)
  {
    System.err.println("snark: " + s
		       + ((debug < INFO|| t == null) ? "" : (": " + t)));
    if (debug >= INFO && t != null)
      t.printStackTrace();
    System.exit(-1);
  }

  /**
   * Show debug info if debug is true.
   */
  public static void debug(String s, int level)
  {
    if (debug >= level)
      System.out.println(s);
  }
}
